/*
 *    GDMcommunication routines
 *    (c)2001 Queen of England, (c)2002,2003 George Lebl
 *    
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *   
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *   
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include "config.h"
#include <libgnome/libgnome.h>
#include <libgnomeui/libgnomeui.h>
#include <gdk/gdkx.h>
#include <X11/Xauth.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <errno.h>

#include <viciousui.h>

#include "gdm.h"
#include "gdmdynamic.h"

static gboolean debug = FALSE;

static char *
do_command (int fd, const char *command, gboolean get_response)
{
	GString *str;
	char buf[1];
	char *cstr;
	int ret;
#ifndef MSG_NOSIGNAL
	void (*old_handler)(int);
#endif

	if (debug)
		fprintf(stderr, "Sending command: '%s'\n", command);

	cstr = g_strdup_printf ("%s\n", command);

#ifdef MSG_NOSIGNAL
	ret = send (fd, cstr, strlen (cstr), MSG_NOSIGNAL);
#else
	old_handler = signal (SIGPIPE, SIG_IGN);
	ret = send (fd, cstr, strlen (cstr), 0);
	signal (SIGPIPE, old_handler);
#endif

	if (ret < 0)
		return NULL;

	if ( ! get_response)
		return NULL;

	str = g_string_new (NULL);
	while (read (fd, buf, 1) == 1 &&
	       buf[0] != '\n') {
		g_string_append_c (str, buf[0]);
	}

	if (debug)
		fprintf(stderr, "  Got response: '%s'\n", str->str);

	cstr = str->str;
	g_string_free (str, FALSE);
	return cstr;
}

static gboolean
version_ok_p (const char *version, const char *min_version)
{
	int a = 0, b = 0, c = 0, d = 0;
	int mina = 0, minb = 0, minc = 0, mind = 0;

	/* Note that if some fields in the version don't exist, then
	 * we don't mind, they are zero */
	sscanf (version, "%d.%d.%d.%d", &a, &b, &c, &d);
	sscanf (min_version, "%d.%d.%d.%d", &mina, &minb, &minc, &mind);

	if ((a > mina) ||
	    (a == mina && b > minb) ||
	    (a == mina && b == minb && c > minc) ||
	    (a == mina && b == minb && c == minc && d >= mind))
		return TRUE;
	else
		return FALSE;
}

char *
gdmdynamic_call_gdm (const char *command, const char * param, const char *min_version, int tries)
{
	struct sockaddr_un addr;
	int fd;
	char *ret;
	char *cstr;

	if (tries <= 0)
		return NULL;


	fd = socket (AF_UNIX, SOCK_STREAM, 0);
	if (fd < 0) {
		sleep(5);
		return gdmdynamic_call_gdm (command, param, min_version, tries - 1);
	}

	strcpy (addr.sun_path, GDM_DYN_SOCKET);
	addr.sun_family = AF_UNIX;

	if (connect (fd, (struct sockaddr *)&addr, sizeof (addr)) < 0) {
		close (fd);
		sleep(5);
		return gdmdynamic_call_gdm (command, param, min_version, tries - 1);
	}


	/* Version check first */
	ret = do_command (fd, GDM_SUP_VERSION, TRUE /* get_response */);
	if (ret == NULL) {
		close (fd);
		sleep(1);
		return gdmdynamic_call_gdm (command, param, min_version, tries - 1);
	}
	if (strncmp (ret, "GDM ", strlen ("GDM ")) != 0) {
		g_free (ret);
		close (fd);
		return NULL;
	}
	if ( ! version_ok_p (&ret[4], min_version)) {
		g_free (ret);
		do_command (fd, GDM_SUP_CLOSE, FALSE /* get_response */);
		close (fd);
		return NULL;
	}
	g_free (ret);

	if (param && *param != 0)
		cstr = g_strdup_printf ("%s %s", command, param);
	else
		cstr = g_strdup(command);

	ret = do_command (fd, cstr, TRUE /* get_response */);

	g_free(cstr);

	if (ret == NULL) {
		close (fd);
		sleep(1);
		return gdmdynamic_call_gdm (command, param, min_version, tries - 1);
	}

	do_command (fd, GDM_SUP_CLOSE, FALSE /* get_response */);

	close (fd);

	return ret;
}



gboolean
gdmdynamic_check (gboolean gui_bitching)
{
	FILE *fp = NULL;
	long pid;
	char *pidfile;
	struct stat s;
	int statret;

	pidfile = ve_config_get_string (ve_config_get (GDM_CONFIG_FILE),
					GDM_KEY_PIDFILE);

	pid = 0;
	if (pidfile != NULL)
		fp = fopen (pidfile, "r");
	if (fp != NULL) {
		fscanf (fp, "%ld", &pid);
		fclose (fp);
	}

	g_free (pidfile);

	errno = 0;
	if (pid <= 1 ||
	    (kill (pid, 0) < 0 &&
	     errno != EPERM)) {
		return FALSE;
	}

	IGNORE_EINTR (statret = stat (GDM_DYN_SOCKET, &s));
	if (statret < 0 ||
	    s.st_uid != 0 ||
	    access (GDM_DYN_SOCKET, R_OK|W_OK) != 0) {
		return FALSE;
	}

	return TRUE;
}


void
gdmdynamic_set_debug (gboolean enable)
{
	debug = enable;
}

int
main (int argc, char *argv[])
{
	char *command = NULL;
	char *version;
	char *ret = NULL;
	char *params = NULL;
	int optc;
	int verbose = 0;
	int background = 0;


	version = "2.2.4.0";

	while ((optc = getopt(argc, argv, "a:d:r:blv")) != EOF) {
		switch (optc) {
		case 'a':
			command = GDM_SUP_ADD_X_SERVER;
			params = optarg;
			break;

		case 'b':
			background = 1;
			break;

		case 'd':
			command = GDM_SUP_REMOVE_X_SERVER;
			params = optarg;
			break;

		case 'l':
			command = GDM_SUP_LIST_DISPLAYS;
			break;

		case 'r':
			command = GDM_SUP_RELEASE_X_SERVERS;
			params = optarg;
			break;

		case 'v':
			verbose++;
			break;
		}
	}

	/* handle background option by forking and exiting the parent */

	if (background) {
		if ((background = fork()) != 0)  {
			if (background < 0)
				exit(1);
			else
				exit(0);
		}
	}

	gdmdynamic_set_debug (verbose);


	ret = gdmdynamic_call_gdm (command, params, version, 5);
	if (ret != NULL &&
	    strncmp (ret, "OK", 2) == 0) {

		if (strcmp(command, GDM_SUP_LIST_DISPLAYS) == 0) {
			ret += 2;
			printf("%s\n", ret);
		}

		/* all fine and dandy */
		return 0;
	}

	if (verbose)
		fprintf(stderr,"%s\n", ret);


	return 1;
}

